/*
 * segmentation.c
 *
 * Created on: Sep 9, 2013
 * 			Author: Amir Yazdanbakhsh <a.yazdanbakhsh@gatech.edu>
 */


#include "segmentation.h"
//#include <stdio.h>
//#include <stdlib.h>
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>

using namespace std;

void RgbPixel::Set_RgbPixel (int E, int M) {

	//cout << "E = " << E << "M = " << M << endl;
	Exp = E;
	Mant = M;
	r = new flx::floatxr<> (Exp, Mant);
	g = new flx::floatxr<> (Exp, Mant);
	b = new flx::floatxr<> (Exp, Mant);
	distance = new flx::floatxr<> (Exp, Mant);



}


RgbPixel::RgbPixel () {

	Exp = 0;
	Mant = 0;
	r = NULL; 
	g = NULL; 
	b = NULL; 
	distance = NULL; 
	cluster = 0;

	
}


RgbPixel::~RgbPixel () {

	delete (r);
	delete (g);
	delete (b);
	delete (distance);
}








int count = 0;
#define MAX_COUNT 1200000


RgbImage::RgbImage ()  {

  w = 0;
  h = 0;
  pixels = NULL;
  meta = NULL;


}
RgbImage::~RgbImage () {
  int i;

  if (meta != NULL)
          delete(meta);

  if (pixels == NULL)
          return;

  for (i = 0; i < h; i++)
          if (pixels != NULL)
                  delete [] pixels[i];

  delete [] pixels;
}

int readCell(FILE *fp, char* w) {
  int c;
  int i;

  w[0] = 0;
  for (c = fgetc(fp), i = 0; c != EOF; c = fgetc(fp)) {
    if (c == ' ' || c == '\t') {
      if (w[0] != '\"')
              continue;
    }

    if (c == ',' || c == '\n') {
      if (w[0] != '\"')
        break;
      else if (c == '\"') {
        w[i] = c;
        i++;
        break;
      }
    }

    w[i] = c;
    i++;
  }
  w[i] = 0;

  return c;
}

int RgbImage::loadRgbImage(const char* fileName,  float scale, int Exp, int Mant ) {
  int c;
  int i;
  int j;
  char w_str[256];
  FILE *fp;

  //printf("Loading %s ...\n", fileName);

  fp = fopen(fileName, "r");
  if (!fp) {
    printf("Warning: Oops! Cannot open %s!\n", fileName);
    return 0;
  }

  c = readCell(fp, w_str);
  w = atoi(w_str);
  c = readCell(fp, w_str);
  h = atoi(w_str);

  //printf("%d x %d\n", w, h);

  pixels = new RgbPixel*[h]; 

  if (pixels == NULL) {
    printf("Warning: Oops! Cannot allocate memory for the pixels!\n");

    fclose(fp);

    return 0;
  }
  
  c = 0;
  for(i = 0; i < h; i++) {
    pixels[i] =  new RgbPixel[w]; 
    if (pixels[i] == NULL) {
      c = 1;
      break;
    }
    for(j = 0; j < w; j++) 
        pixels[i][j].Set_RgbPixel (Exp,Mant); 

  }

  if (c == 1) {
    printf("Warning: Oops! Cannot allocate memory for the pixels!\n");

    for (i--; i >= 0; i--)
       delete (pixels[i]);
    delete(pixels);

    fclose(fp);

    return 0;
  }

  for(i = 0; i < h; i++) {
    for(j = 0; j < w; j++) {
      c = readCell(fp, w_str);
      *(pixels[i][j].r) = atoi(w_str) / scale;

      c = readCell(fp, w_str);
      *(pixels[i][j].g) = atoi(w_str) / scale;

      c = readCell(fp, w_str);
      *(pixels[i][j].b) = atoi(w_str) / scale;

      pixels[i][j].cluster = 0;
      *(pixels[i][j].distance) = 0.;
    }
  }

  c = readCell(fp, w_str);
  meta = new char[strlen(w_str)];  
  if(meta == NULL) {
    printf("Warning: Oops! Cannot allocate memory for the pixels!\n");

    for (i = 0; i < h; i++)
            delete(pixels[i]);
    delete(pixels);

    fclose(fp);

    return 0;

  }
  strcpy(meta, w_str);

  //printf("%s\n", image->meta);

  //printf("w=%d x h=%d\n", image->w, image->h);

  return 1;
}

int RgbImage::saveRgbImage(const char* fileName, float scale) {
  int i;
  int j;
  FILE *fp;

  //printf("Saving %s ...\n", fileName);

  fp = fopen(fileName, "w");
  if (!fp) {
    printf("Warning: Oops! Cannot open %s!\n", fileName);
    return 0;
  }

  fprintf(fp, "%d,%d\n", w, h);
  //printf("%d,%d\n", image->w, image->h);

  for(i = 0; i < h; i++) {
    for(j = 0; j < (w - 1); j++) {
      fprintf(fp, "%d,%d,%d,", (int)(pixels[i][j].getr() * scale) , (int)(pixels[i][j].getg() * scale), (int)(pixels[i][j].getb() * scale));
    }
    fprintf(fp, "%d,%d,%d\n", (int)(pixels[i][j].getr() * scale), (int)(pixels[i][j].getg() * scale), (int)(pixels[i][j].getb() * scale));
  }

  fprintf(fp, "%s", meta);
  //printf("%s\n", image->meta);

  fclose(fp);

  return 1;
}


void RgbImage::makeGrayscale() {
  int i;
  int j;
  float luminance;

  float rC = 0.30;
  float gC = 0.59;
  float bC = 0.11;

  for(i = 0; i < h; i++) {
    for(j = 0; j < w; j++) {
      luminance =
              rC * pixels[i][j].getr() +
              gC * pixels[i][j].getg() +
              bC * pixels[i][j].getb();

      *(pixels[i][j].r) = (unsigned char)luminance;
      *(pixels[i][j].g) = (unsigned char)luminance;
      *(pixels[i][j].b) = (unsigned char)luminance;
    }
  }
}


void Centroid::Set_Centroid (int E, int M) {

	Exp = E;
	Mant = M;
	r = new flx::floatxr<> (Exp, Mant);
	g = new flx::floatxr<> (Exp, Mant);
	b = new flx::floatxr<> (Exp, Mant);


}

Centroid::Centroid () {

	Exp = 0;
	Mant = 0;
	r = NULL;
	g = NULL;
	b = NULL;
  	n = 0;

}

Centroid::~Centroid () {

	delete r;
	delete g;
	delete b;

}






float euclideanDistance(RgbPixel* p, Centroid* c1) {
  float r;

  r = 0;
  double r_tmp;

  double dataIn[6];
  dataIn[0] = p->getr();
  dataIn[1] = p->getg ();
  dataIn[2] = p->getb();
  dataIn[3] = c1->getr();
  dataIn[4] = c1->getg();
  dataIn[5] = c1->getb();


  r += (p->getr() - c1->getr()) * (p->getr() - c1->getr());
  r += (p->getg() - c1->getg()) * (p->getg() - c1->getg());
  r += (p->getb() - c1->getb()) * (p->getb() - c1->getb());

  r_tmp = sqrt(r);


  return r_tmp;
}

int pickCluster(RgbPixel* p, Centroid* c1) {
  float d1;

  d1 = euclideanDistance(p, c1);

  if (*(p->distance) <= d1)
    return 0;

  return 1;
}

void assignCluster(RgbPixel* p, Clusters* clusters) {
  int i = 0;

  float d;
  d = euclideanDistance(p, &clusters->centroids[i]);
  *(p->distance) = d;

  for(i = 1; i < clusters->k; ++i) {
    d = euclideanDistance(p, &clusters->centroids[i]);
    if (d < *(p->distance)) {
      p->cluster = i;
      *(p->distance) = d;
    }
  }
}

int initClusters(Clusters* clusters, int k, float scale, int Exp, int Mant ) {
  int i;
  float x;

  //clusters->centroids = (Centroid*)malloc(k * sizeof(Centroid));
  clusters->centroids = new Centroid[k];
  for (i = 0; i < k; i ++)  
	clusters->centroids[i].Set_Centroid (Exp, Mant);


  if (clusters == NULL) {
    printf("Warning: Oops! Cannot allocate memory for the clusters!\n");

    return 0;
  }

  clusters->k = k;
  for (i = 0; i < clusters->k; i ++) {
    x = (((float)rand())/RAND_MAX) * scale;
    *(clusters->centroids[i].r) = x;

    x = (((float)rand())/RAND_MAX) * scale;
    *(clusters->centroids[i].g) = x;

    x = (((float)rand())/RAND_MAX) * scale;
    *(clusters->centroids[i].b) = x;

    clusters->centroids[i].n = 0;
  }

  return 1;
}

void freeClusters(Clusters* clusters) {
  if (clusters != NULL) {
	if (clusters->centroids != NULL) {
	
		delete [] clusters->centroids;
	}
 }
    
}




void segmentImage(RgbImage* image, Clusters* clusters, int n) {
  int i;
  int x, y;
  int c;

  for (i = 0; i < n; ++i) {
    for (y = 0; y < image->h; y++) {
      for (x = 0; x < image->w; x++) {
              assignCluster(&image->pixels[y][x], clusters);
      }
    }

    /** Recenter */
    for (c  = 0; c < clusters->k; ++c) {
      *(clusters->centroids[c].r) = 0.;
      *(clusters->centroids[c].g) = 0.;
      *(clusters->centroids[c].b) = 0.;
      clusters->centroids[c].n = 0;
    }
    for (y = 0; y < image->h; y++) {
      for (x = 0; x < image->w; x++) {
        *(clusters->centroids[image->pixels[y][x].cluster].r) += *(image->pixels[y][x].r);
        *(clusters->centroids[image->pixels[y][x].cluster].g) += *(image->pixels[y][x].g);
        *(clusters->centroids[image->pixels[y][x].cluster].b) += *(image->pixels[y][x].b);
        clusters->centroids[image->pixels[y][x].cluster].n += 1;
      }
    }
    for (c  = 0; c < clusters->k; ++c) {
      if (clusters->centroids[c].n != 0) {
        *(clusters->centroids[c].r) = clusters->centroids[c].getr() / clusters->centroids[c].n;
        *(clusters->centroids[c].g) = clusters->centroids[c].getg() / clusters->centroids[c].n;
        *(clusters->centroids[c].b) = clusters->centroids[c].getb() / clusters->centroids[c].n;
      }
    }
  }

  for (y = 0; y < image->h; y++) {
    for (x = 0; x < image->w; x++) {
      *(image->pixels[y][x].r) = clusters->centroids[image->pixels[y][x].cluster].getr();
      *(image->pixels[y][x].g) = clusters->centroids[image->pixels[y][x].cluster].getg();
      *(image->pixels[y][x].b) = clusters->centroids[image->pixels[y][x].cluster].getb();
    }
  }
}
