/*
 * segmentation.h
 * 
 * Created on: Sep 9, 2013
 * 			Author: Amir Yazdanbakhsh <a.yazdanbakhsh@gatech.edu>
 */


#ifndef SEGMENTATION_H_
#define SEGMENTATION_H_

#include <string>
#include <iostream>

using namespace std;

#include "floatx.hpp"




class RgbPixel {

public:
	int Exp, Mant;
	flx::floatxr<> *r;
	flx::floatxr<> *g;
	flx::floatxr<> *b;
	int cluster;
	flx::floatxr<> *distance;
	RgbPixel ();
	void Set_RgbPixel (int E, int M);
	~RgbPixel ();

	flx::floatxr<> getr() {
		return *r;
	};
	flx::floatxr<> getg() {
		return *g;
	};
	flx::floatxr<> getb() {
		return *b;
	};
};

class RgbImage {
public:
   int w;
   int h;
   RgbPixel** pixels;
   char* meta;

   RgbImage ();
   ~RgbImage ();

   int loadRgbImage(const char* fileName, float scale, int Exp, int Mant );
   int saveRgbImage(const char* fileName, float scale);
   void makeGrayscale();
   
};

class Centroid {

public:
   int Exp, Mant;
   flx::floatxr<> *r;
   flx::floatxr<> *g;
   flx::floatxr<> *b;
   int n;


   void Set_Centroid (int E, int M);
   Centroid ();
   ~Centroid ();

	flx::floatxr<> getr() {
		return *r;
	};
	flx::floatxr<> getg() {
		return *g;
	};
	flx::floatxr<> getb() {
		return *b;
	};


} ;

typedef struct {
   int k;
   Centroid* centroids;
} Clusters;


int readCell(FILE *fp, char* w);

int initClusters(Clusters* clusters, int k, float scale, int Exp, int Mant );
void segmentImage(RgbImage* image, Clusters* clusters, int n);
void freeClusters(Clusters* clusters);

#endif /* SEGMENTATION_H_ */
